# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import httplib
from networking_zte.common import servermanager
from networking_zte.ml2.common.rest.ml2_zenic import config as pl_config
from neutron.db import common_db_mixin as base_db
try:
    from neutron_lib.api.definitions import port_security as psec
except Exception:
    from neutron.extensions import portsecurity as psec
from neutron.plugins.ml2 import driver_api as api
from oslo_log import log
from oslo_serialization import jsonutils


LOG = log.getLogger(__name__)

# The following are used to invoke the API on the external controller
TENANT = 'tenant'
NETWORK = 'network'
SUBNET = 'subnet'
PORT = 'port'
ROUTER = 'router'
FLOATING_IP = 'floating-ip'
VXLAN_TUNNEL = 'vxlan-tunnel'
SECURITY_GROUP = 'sg'
SECURITY_GROUP_RULE = 'sg-rule'
CLASSIFIER = 'classifier'

QOS_POLICY = 'qos-policy'
QOS_BANDWIDTH_RULE = 'qos-bandwidth-limit-rule'

BASE_URI = '/restconf/operations/zenic-vdcapp-model:'
SUCCESS_CODES = range(200, 207)
FAILURE_CODES = [
    0,
    9,
    301,
    302,
    303,
    400,
    401,
    403,
    404,
    409,
    500,
    501,
    502,
    503,
    504,
    505]


class ZenicServerPool(servermanager.ServerPool, base_db.CommonDbMixin):
    """Zenic Server Pool for Zenic Mechanism Driver.
    This server pool has the network,subnet,port and security group operations
    of create, update and delete, to the Zenic Controller.
    """

    def __init__(
            self,
            servers,
            auth,
            zenic_version,
            enable_qos,
            ssl,
            no_ssl_validation,
            ssl_sticky,
            ssl_cert_directory,
            consistency_interval,
            timeout=False,
            cache_connections=True,
            base_uri=BASE_URI,
            success_codes=SUCCESS_CODES,
            failure_codes=FAILURE_CODES,
            name='ZenicRestProxy'):
        super(ZenicServerPool, self).__init__(
            servers, auth, ssl, no_ssl_validation, ssl_sticky,
            ssl_cert_directory, consistency_interval, timeout,
            cache_connections, base_uri, success_codes, failure_codes, name)
        version = zenic_version.split('.')
        version = version[0] + version[1]
        if (version.isdigit() is False) or (int(version) < 403):
            LOG.error(_("zenic_version error! zenic_version = %s"), version)
        self.zenic_version = int(version)
        self.enable_qos = enable_qos
        pl_config.register_config()
        self.enable_mitaka_qos = pl_config.cfg.CONF.RESTPROXY.enable_M_qos
        self.enable_bandwidth = pl_config.cfg.CONF.RESTPROXY.enable_bandwidth
        self.flat_segment_id = pl_config.cfg.CONF.RESTPROXY.flat_segment_id
        self.vlan_transparent = pl_config.cfg.CONF.RESTPROXY.vlan_transparent
        self.hierarchical = pl_config.cfg.CONF. \
            RESTPROXY.enable_hierarchical_port
        self.en_security_group = False
        LOG.info(_("proxyMechanismDriver enable_mitaka_qos = %s"),
                 self.enable_mitaka_qos)
        LOG.info(_("proxyMechanismDriver enable_qos = %s"), self.enable_qos)

    @staticmethod
    def validate_dict(instance, key, default_val):
        return instance[key] if (key in instance and
                                 instance[key] is not None) else default_val

    @staticmethod
    def validate_ipv4(ip_in, default_val):
        return ip_in if (ip_in != 0 and (ip_in is not None)) else default_val

    def construct_network_info(self, mech_context, action):
        network = mech_context.current
        LOG.debug("construct_network_info network = %s" % network)

        if action == 'DELETE' or action == 'GET':
            network_info = {
                "input": {
                    "id": network['id']
                }
            }
        else:
            network_info = {
                "input": {
                    "id": network['id'],
                    "name": network['name'],
                    "admin_state_up": network['admin_state_up'],
                    "tenant_id": network['tenant_id'],
                    "shared": network['shared'],
                    "band_width": self.validate_dict(network, 'bandwidth', 0),
                    "burst_size": self.validate_dict(network, 'cbs', 0),
                    "dscp": self.validate_dict(network, 'dscp', 0),
                    "external":
                        self.validate_dict(network, 'router:external', False),
                }
            }

            input = network_info['input']
            if self.vlan_transparent == 'True':
                if network['vlan_transparent'] is not None:
                    input['vlan_transparent'] = \
                        self.validate_dict(network, 'vlan_transparent', False)
            if network.get('provider:segmentation_id', None) is not None:
                if network.get('provider:network_type') != "flat":
                    input['segmentation_id'] =\
                        mech_context.network_segments[0]['segmentation_id']
                else:
                    input['segmentation_id'] = self.flat_segment_id
            else:
                return None

            if self.zenic_version > 403:
                if 'mtu' in network:
                    input['mtu'] = self.validate_dict(network, "mtu", 0)
                if self.enable_mitaka_qos == "True":
                    if 'qos_policy_id' in network.keys():
                        input['qos_policy_id'] = network['qos_policy_id']

        if action == 'POST':
            input = network_info['input']
            if self.zenic_version > 403:
                if 'port_security_enabled' in network:
                    if self.en_security_group:
                        input['port_security_enabled'] = \
                            self.validate_dict(network, psec.PORTSECURITY,
                                               True)

        return network_info

    def construct_subnet_info(self, mech_context, action):
        subnet = mech_context.current
        LOG.debug("construct_subnet_info subnet = %s" % subnet)

        if action == 'DELETE' or action == 'GET':
            subnet_info = {
                "input": {
                    "id": subnet['id']
                }
            }
        else:
            if subnet['ip_version'] == 6:
                gateway_default = "::"
            else:
                gateway_default = "0.0.0.0"

            subnet_info = {
                "input": {
                    "id": subnet['id'],
                    "subnet_name": subnet['name'],
                    # "network_id": subnet['network_id'],
                    # "tenant_id": tenant_id,
                    "dns_nameservers": ','.join(subnet['dns_nameservers']),
                    "allocation_pools": subnet['allocation_pools'],
                    "host_routes":
                        '\r\n'.join(','.join([route.get("destination", ""),
                                              route.get("nexthop", "")])
                                    for route in self.validate_dict(
                            subnet, 'host_routes', [])),
                    # "ip_version": subnet['ip_version'],
                    "gateway_ip": self.validate_ipv4(
                        subnet['gateway_ip'], gateway_default),
                    # "cidr": subnet['cidr']
                }
            }

        if action == 'POST':
            input = subnet_info['input']
            input['network_id'] = subnet['network_id']
            input['tenant_id'] = subnet['tenant_id']
            input['cidr'] = subnet['cidr']
            input['ip_version'] = subnet['ip_version']

        return subnet_info

    def get_next_segment_id(self, action, mech_context, host_id):
        next_segment = mech_context.bottom_bound_segment
        top_segment = mech_context.top_bound_segment
        LOG.info(_("port next_segment=%s"), next_segment)
        LOG.info(_("port top_segment=%s"), top_segment)
        if action == 'PUT':
            if next_segment is not None and top_segment is not None\
                    and host_id != 'unknow':
                return next_segment['segmentation_id']
        else:
            return 999999
        return None

    def update_port_filter(self, current_port, original_port):
        if current_port is None or original_port is None:
            return True
        if current_port.keys() != original_port.keys():
            return True
        for key in original_port.keys():
            if key != 'updated_at' and key != 'revision_number' \
                    and key != 'status' and original_port[key] != \
                    current_port[key]:
                return True
        return False

    def construct_port_info(self, mech_context, action):
        current_port = mech_context.current
        LOG.debug("construct_port_info current_port = %s" % current_port)
        original_port = mech_context.original
        LOG.debug("construct_port_info original_port = %s" % original_port)
        flag = self.update_port_filter(current_port, original_port)
        if not flag:
            LOG.debug("update port filter")
            return None
        else:
            port = current_port

        if action == 'DELETE' or action == 'GET':
            port_info = {
                "input": {
                    "id": port["id"]
                }
            }
        else:
            port_info = {
                "input": {
                    "id": port['id'],
                    "name": port['name'],
                    "allowed_address_pairs":
                        [{'ip_address': pairs['ip_address'],
                          'mac_address': pairs['mac_address']}
                         for pairs in port['allowed_address_pairs']],
                    "admin_state_up": port["admin_state_up"],
                    # "network_id": port["network_id"],
                    # "tenant_id": tenant_id,
                    # "mac_address": port["mac_address"],
                    # "binding_profile": str(port['binding:profile']),
                    "device_owner": port["device_owner"],
                    "fixed_ips": [{'subnet_id': ip["subnet_id"],
                                   'ip_address': ip["ip_address"]}
                                  for ip in port["fixed_ips"]],
                    "burst_size": self.validate_dict(port, 'cbs', 0),
                    "vnic_type": self.validate_dict(port,
                                                    'binding:vnic_type',
                                                    'normal'),
                    "vif_type": self.validate_dict(port,
                                                   'binding:vif_type',
                                                   'unbound'),
                    "host_id": self.validate_dict(port,
                                                  'binding:host_id', ''),
                }
            }

        if action == 'POST' or action == 'PUT':
            input = port_info['input']
            if self.zenic_version > 403:
                if 'extra_dhcp_opts' in port:
                    input['extra_dhcp_opts'] = [
                        {'opt_value': dhcp["opt_value"],
                         'ip_version': dhcp["ip_version"],
                         'opt_name': dhcp["opt_name"]}
                        for dhcp in port["extra_dhcp_opts"]]
                if self.enable_qos == "True":
                    if port.get('qos', None):
                        input['qos_policy_id'] = port['qos']
                    else:
                        input['qos_policy_id'] = ''
                if self.enable_mitaka_qos == "True":
                    if 'qos_policy_id' in port.keys():
                        input['qos_policy_id'] = port['qos_policy_id']

                pl_config.register_config()
                if pl_config.cfg.CONF.RESTPROXY.enable_hierarchical_port == \
                        "True":
                    next_segment = self.get_next_segment_id(
                        action, mech_context, port['binding:host_id'])
                    if next_segment:
                        input['next_segment_to_bind'] = next_segment
                if original_port:
                    if port['binding:profile'] != {} \
                            or original_port['binding:profile'] != {}:
                        input["binding_profile"] = jsonutils.dumps(
                            port['binding:profile'])
                else:
                    if port['binding:profile'] != {}:
                        input["binding_profile"] = jsonutils.dumps(
                            port['binding:profile'])
                if self.enable_bandwidth == "True":
                    input["band_width"] = self.validate_dict(port, 'bandwidth',
                                                             0)
                    input["dscp"] = self.validate_dict(port, 'dscp', 0)
                if self.en_security_group:
                    if 'port_security_enabled' in port.keys():
                        if port['port_security_enabled'] is True:
                            input["security_groups"] = port["security_groups"]
                        else:
                            input["security_groups"] = []
                    else:
                        input["security_groups"] = port["security_groups"]
                else:
                    input["security_groups"] = []

        if action == 'POST':
            input = port_info['input']
            input['network_id'] = port['network_id']
            input['tenant_id'] = port['tenant_id']
            input['mac_address'] = port['mac_address']
            if self.zenic_version > 403:
                if 'port_security_enabled' in port and self.en_security_group:
                    input['port_security_enabled'] = \
                        self.validate_dict(port, psec.PORTSECURITY, True)
        return port_info

    def construct_security_group_info(self, mech_context, action):
        sg = mech_context.current
        LOG.debug("construct_security_group_info sg = %s" % sg)

        if action == 'DELETE' or action == 'GET':
            security_group_info = {"input": {"id": sg["id"]}}
        elif action == 'PUT':
            security_group_info = {
                "input": {
                    "id": sg['id'],
                    "name": sg['name'],
                    "description": sg["description"],
                }
            }
        else:
            security_group_info = {
                "input": {
                    "id": sg['id'],
                    "name": sg['name'],
                    "description": sg["description"],
                    "tenant_id": sg['tenant_id']
                }
            }

        if action == "POST":
            security_group_rules = self.validate_dict(
                sg, 'security_group_rules', None)
            if security_group_rules is not None:
                sg_rules = []
                for rule in security_group_rules:
                    ether_type = self.validate_dict(rule, 'ethertype', None)
                    ipv4 = None
                    ipv6 = None
                    if ether_type and ether_type.find('4') != -1:
                        ipv4 = self.validate_dict(
                            rule, 'remote_ip_prefix', None)
                    elif ether_type and ether_type.find('6') != -1:
                        ipv6 = self.validate_dict(
                            rule, 'remote_ip_prefix', None)
                    else:
                        LOG.error("ether_type:%s is error!" % ether_type)

                    sg_rule = {
                        "id": rule['id'],
                        "port_range_max":
                            self.validate_dict(rule, 'port_range_max', 0),
                        "port_range_min":
                            self.validate_dict(rule, 'port_range_min', 0),
                        "protocol":
                            self.validate_dict(rule, 'protocol', None),
                        "remote_group_id":
                            self.validate_dict(rule, 'remote_group_id', None),
                        "remote_ipv4_prefix": ipv4,
                        "remote_ipv6_prefix": ipv6,
                        "direction":
                            self.validate_dict(rule, 'direction', None),
                        "ethertype": ether_type,
                        "tenant_id": sg['tenant_id'],
                        "security_group_id":
                            self.validate_dict(rule, 'security_group_id', None)
                    }
                    sg_rules.append(sg_rule)
                security_group_info['input']['security_group_rules'] =\
                    sg_rules
        return security_group_info

    def construct_security_group_rule_info(self, mech_context, action):
        rule = mech_context.current
        LOG.debug("construct_security_group_rule_info rule = %s" % rule)

        ether_type = self.validate_dict(rule, 'ethertype', None)
        ipv4 = None
        ipv6 = None
        if ether_type and ether_type.find('4') != -1:
            ipv4 = self.validate_dict(rule, 'remote_ip_prefix', None)
        elif ether_type and ether_type.find('6') != -1:
            ipv6 = self.validate_dict(rule, 'remote_ip_prefix', None)
        else:
            LOG.error("ether_type:%s is error!" % ether_type)

        if action == 'DELETE' or action == 'GET':
            security_group_rule_info = {"input": {"id": rule["id"]}}
        else:
            security_group_rule_info = {
                "input": {
                    "id": rule['id'],
                    "port_range_max":
                        self.validate_dict(rule, 'port_range_max', 0),
                    "port_range_min":
                        self.validate_dict(rule, 'port_range_min', 0),
                    "protocol":
                        self.validate_dict(rule, 'protocol', None),
                    "remote_group_id":
                        self.validate_dict(rule, 'remote_group_id', None),
                    "remote_ipv4_prefix": ipv4,
                    "remote_ipv6_prefix": ipv6,
                    "security_group_id":
                        self.validate_dict(rule, 'security_group_id', None)
                }
            }

        if action == 'POST':
            input = security_group_rule_info['input']
            input['direction'] = self.validate_dict(rule, 'direction', None)
            input['ethertype'] = ether_type
            input['tenant_id'] = rule['tenant_id']

        return security_group_rule_info

    def set_enable_security_group(self, en_security_group):
        self.en_security_group = en_security_group
        LOG.info(_(
            "zenic_l2restconf en_security_group = %s"),
            self.en_security_group)

    def rest_create_tenant(self, tenant_id, tenant_name, description):
        tenant_data = {"id": tenant_id,
                       "name": tenant_name,
                       "description": description}
        data = {"input": tenant_data}
        resource = 'add-' + TENANT
        err_str = _("Unable to create tenant: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_update_tenant(self, tenant_id, tenant_name, description):
        tenant_data = {"id": tenant_id,
                       "name": tenant_name,
                       "description": description}
        data = {"input": tenant_data}
        resource = 'update-' + TENANT
        err_str = _("Unable to update tenant: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_delete_tenant(self, tenant_id):
        tenant_data = {"id": tenant_id}
        data = {"input": tenant_data}
        resource = 'del-' + TENANT
        err_str = _("Unable to delete tenant: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_get_tenant(self, tenant_id):
        tenant_data = {"id": tenant_id}
        data = {"input": tenant_data}
        resource = 'get-' + TENANT
        err_str = _("Unable to get tenant: %s")
        return self.rest_action('POST', resource, data, err_str)

    def rest_create_network(self, mech_context):
        data = self.construct_network_info(mech_context, 'POST')
        if not data:
            return None
        resource = 'add-' + NETWORK
        err_str = _("Unable to create remote network: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_update_network(self, mech_context):
        data = self.construct_network_info(mech_context, 'PUT')
        if not data:
            return None
        resource = 'update-' + NETWORK
        err_str = _("Unable to update remote network: %s")

        ret = self.rest_action('POST', resource, data, err_str)
        if ret[0] == httplib.NOT_FOUND:
            self.rest_create_network(mech_context)

    def rest_delete_network(self, mech_context):
        try:
            from neutron.db import segments_db as db
        except Exception:
            from neutron.plugins.ml2 import db
        session = mech_context._plugin_context.session
        network_id = mech_context.current['id']
        if self.hierarchical == "True":
            dynamic_segment = db.get_dynamic_segment(session, network_id)
            LOG.info(_("delete_network dynamic_segment = %s"), dynamic_segment)
            if dynamic_segment:
                db.delete_network_segment(session,
                                          dynamic_segment[api.SEGMENTATION_ID])
        data = self.construct_network_info(mech_context, 'DELETE')
        resource = 'del-' + NETWORK
        err_str = _("Unable to delete remote network: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_get_network(self, mech_context):
        data = self.construct_network_info(mech_context, 'GET')
        resource = 'get-' + NETWORK
        err_str = _("Unable to get remote network: %s")
        return self.rest_action('POST', resource, data, err_str)

    def rest_create_subnet(self, mech_context):
        data = self.construct_subnet_info(mech_context, 'POST')
        resource = 'add-' + SUBNET
        err_str = _("Unable to create remote subnet: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_update_subnet(self, mech_context):
        data = self.construct_subnet_info(mech_context, 'PUT')
        resource = 'update-' + SUBNET
        err_str = _("Unable to update remote subnet: %s")
        ret = self.rest_action('POST', resource, data, err_str)
        if ret[0] == httplib.NOT_FOUND:
            self.rest_create_subnet(mech_context)

    def rest_delete_subnet(self, mech_context):
        data = self.construct_subnet_info(mech_context, 'DELETE')
        resource = 'del-' + SUBNET
        err_str = _("Unable to delete remote subnet: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_get_subnet(self, mech_context):
        data = self.construct_subnet_info(mech_context, 'GET')
        resource = 'get-' + SUBNET
        err_str = _("Unable to get remote subnet: %s")
        return self.rest_action('POST', resource, data, err_str)

    def rest_create_port(self, mech_context):
        data = self.construct_port_info(mech_context, 'POST')
        resource = 'add-' + PORT
        err_str = _("Unable to create remote port: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_update_port(self, mech_context):
        data = self.construct_port_info(mech_context, 'PUT')
        if not data:
            return None
        resource = 'update-' + PORT
        err_str = _("Unable to update remote port: %s")
        ret = self.rest_action('POST', resource, data, err_str)
        if ret[0] == httplib.NOT_FOUND:
            self.rest_create_port(mech_context)

    def rest_delete_port(self, mech_context):
        data = self.construct_port_info(mech_context, 'DELETE')
        resource = 'del-' + PORT
        err_str = _("Unable to delete remote port: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_get_port(self, mech_context):
        data = self.construct_port_info(mech_context, 'GET')
        resource = 'get-' + PORT
        err_str = _("Unable to get remote port: %s")
        return self.rest_action('POST', resource, data, err_str)

    def rest_create_security_group(self, mech_context):
        data = self.construct_security_group_info(mech_context, 'POST')
        resource = 'add-' + SECURITY_GROUP
        err_str = _("Unable to create remote security_group: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_update_security_group(self, mech_context):
        data = self.construct_security_group_info(mech_context, 'PUT')
        resource = 'update-' + SECURITY_GROUP
        err_str = _("Unable to update remote security_group: %s")
        ret = self.rest_action('POST', resource, data, err_str)
        if ret[0] == httplib.NOT_FOUND:
            self.rest_create_security_group(mech_context)

    def rest_delete_security_group(self, mech_context):
        data = self.construct_security_group_info(mech_context, 'DELETE')
        resource = 'del-' + SECURITY_GROUP
        err_str = _("Unable to delete remote security_group: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_get_security_group(self, mech_context):
        data = self.construct_security_group_info(mech_context, 'GET')
        resource = 'get-' + SECURITY_GROUP
        err_str = _("Unable to get remote security_group: %s")
        return self.rest_action('POST', resource, data, err_str)

    def rest_create_security_group_rule(self, mech_context):
        data = self.construct_security_group_rule_info(mech_context, 'POST')
        resource = 'add-' + SECURITY_GROUP_RULE
        err_str = _("Unable to create remote security_group_rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_update_security_group_rule(self, mech_context):
        data = self.construct_security_group_rule_info(mech_context, 'PUT')
        resource = 'update-' + SECURITY_GROUP_RULE
        err_str = _("Unable to update remote security_group_rule: %s")
        ret = self.rest_action('POST', resource, data, err_str)
        if ret[0] == httplib.NOT_FOUND:
            self.rest_create_security_group_rule(mech_context)

    def rest_delete_security_group_rule(self, mech_context):
        data = self.construct_security_group_rule_info(mech_context, 'DELETE')
        resource = 'del-' + SECURITY_GROUP_RULE
        err_str = _("Unable to delete remote security_group_rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_get_security_group_rule(self, mech_context):
        data = self.construct_security_group_rule_info(mech_context, 'GET')
        resource = 'get-' + SECURITY_GROUP_RULE
        err_str = _("Unable to get remote security_group_rule: %s")
        return self.rest_action('POST', resource, data, err_str)

    @staticmethod
    def construct_qos_policy_info(mech_context, action):
        qos_policy = mech_context.current
        qos_policy_info = {
            "input": {
                "id": qos_policy['id'],
            }
        }
        input = qos_policy_info['input']
        if 'type' in qos_policy:
            input['type'] = qos_policy['type']
        if 'description' in qos_policy:
            input['description'] = qos_policy['description']
        if 'tenant_id' in qos_policy:
            input['tenant_id'] = qos_policy['tenant_id']
        if 'shared' in qos_policy:
            input['shared'] = qos_policy['shared']
        if 'name' in qos_policy:
            input['name'] = qos_policy['name']

        return qos_policy_info

    @staticmethod
    def construct_bandwidth_limit_rule_info(mech_context, action):
        """
        {u'bandwidth_limit_rule':
             {u'max_kbps': u'300',
              u'max_burst_kbps': u'30',
              'tenant_id': u'c414cf57e9fc49efb8e3e0c8d4e41b33',
              'id' :'aa1e84d8-9ea3-45ec-8197-83b7d9abc647'}
        }
        """
        bandwidth_limit_rule = mech_context.current

        bandwidth_rule_info = {
            "input": {
                "id": bandwidth_limit_rule['id'],
                "qos_policy_id": bandwidth_limit_rule['id'],
            }
        }
        input = bandwidth_rule_info['input']
        if 'max_kbps' in bandwidth_limit_rule:
            input['max_kbps'] = bandwidth_limit_rule['max_kbps']
        if 'max_burst_kbps' in bandwidth_limit_rule:
            input['max_burst_kbps'] = bandwidth_limit_rule['max_burst_kbps']

        if 'tenant_id' in bandwidth_limit_rule:
            input['tenant_id'] = bandwidth_limit_rule['tenant_id']
        return bandwidth_rule_info

    @staticmethod
    def construct_delete_qos_info(mech_context):
        qos = mech_context.current
        qos_info = {
            "input": {
                "id": qos['id']
            }
        }
        return qos_info

    def rest_create_qos_policy(self, mech_context):
        data = self.construct_qos_policy_info(mech_context, 'CREATE')
        resource = 'add-' + QOS_POLICY
        err_str = _("Unable to add qos policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_update_qos_policy(self, mech_context):
        data = self.construct_qos_policy_info(mech_context, 'UPDATE')
        resource = 'update-' + QOS_POLICY
        err_str = _("Unable to update qos policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_delete_qos_policy(self, mech_context):
        data = self.construct_delete_qos_info(mech_context)
        resource = 'del-' + QOS_POLICY
        err_str = _("Unable to add qos policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_create_qos_bandwidth_limit_rule(self, mech_context):
        data = self.construct_bandwidth_limit_rule_info(mech_context, 'CREATE')
        resource = 'add-' + QOS_BANDWIDTH_RULE
        err_str = _("Unable to add bandwidth limit rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_update_qos_bandwidth_limit_rule(self, mech_context):
        data = self.construct_bandwidth_limit_rule_info(mech_context, 'UPDATE')
        resource = 'update-' + QOS_BANDWIDTH_RULE
        err_str = _("Unable to update bandwidth limit rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def rest_delete_qos_bandwidth_limit_rule(self, mech_context):
        data = self.construct_delete_qos_info(mech_context)
        resource = 'del-' + QOS_BANDWIDTH_RULE
        err_str = _("Unable to del bandwidth limit rule: %s")
        self.rest_action('POST', resource, data, err_str)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
from networking_zte.ml2 import driver_context
from neutron.extensions import allowedaddresspairs as addr_pair
from neutron.extensions import portbindings
from neutron.plugins.ml2 import driver_context as ml2_context
from oslo_log import log
import six

LOG = log.getLogger(__name__)


def try_del(d, keys):
    """Ignore key errors when deleting from a dictionary."""
    for key in keys:
        try:
            del d[key]
        except KeyError:
            pass


@six.add_metaclass(abc.ABCMeta)
class ResourceFilterBase(object):
    @staticmethod
    @abc.abstractmethod
    def filter_create_attributes(resource, context):
        pass

    @staticmethod
    @abc.abstractmethod
    def filter_create_attributes_with_plugin(resource, plugin, dbcontext):
        pass

    @abc.abstractmethod
    def filter_create_attributes_without_plugin(
            resource, dbcontext, resourceplus):
        pass

    @staticmethod
    def _filter_unmapped_null(resource_dict, unmapped_keys):
        keys_to_del = [key for key in unmapped_keys
                       if resource_dict.get(key) is None]
        if keys_to_del:
            try_del(resource_dict, keys_to_del)


class NetworkFilter(ResourceFilterBase):
    _UNMAPPED_KEYS = ['qos_policy_id']

    @classmethod
    def filter_create_attributes(cls, network, context):
        """Filter out network attributes not required for a create."""
        try_del(network, ['status', 'subnets'])
        cls._filter_unmapped_null(network, cls._UNMAPPED_KEYS)

    @classmethod
    def filter_create_attributes_with_plugin(cls, network, plugin, dbcontext):
        context = ml2_context.NetworkContext(plugin, dbcontext, network)
        cls.filter_create_attributes(network, context)
        return context

    @classmethod
    def filter_create_attributes_without_plugin(
            cls, network, dbcontext, networkplus):
        context = ml2_context.NetworkContext(None, dbcontext, network)
        cls.filter_create_attributes(network, context)
        return context


class SubnetFilter(ResourceFilterBase):
    @classmethod
    def filter_create_attributes_with_plugin(cls, subnet, plugin, dbcontext):
        network = plugin.get_network(dbcontext, subnet['network_id'])
        context = ml2_context.SubnetContext(plugin, dbcontext, subnet,
                                            network)
        cls.filter_create_attributes(subnet, context)
        return context

    @classmethod
    def filter_create_attributes_without_plugin(
            cls, subnet, dbcontext, network):
        context = ml2_context.SubnetContext(None, dbcontext, subnet,
                                            network)
        cls.filter_create_attributes(subnet, context)
        return context


class PortFilter(ResourceFilterBase):
    _UNMAPPED_KEYS = ['binding:profile', 'dns_name',
                      'port_security_enabled', 'qos_policy_id']

    @classmethod
    def filter_create_attributes(cls, port, context, network):
        """Filter out port attributes not required for a create."""

        if portbindings.PROFILE not in port:
            port[portbindings.PROFILE] = ''

        cls._filter_unmapped_null(port, cls._UNMAPPED_KEYS)
        try_del(port, ['status'])

        if port['tenant_id'] == '' and 'tenant_id' in network:
            LOG.debug('empty string was passed for tenant_id: %s(port)', port)
            port['tenant_id'] = network['tenant_id']

    @classmethod
    def filter_create_attributes_with_plugin(cls, port, plugin, dbcontext):
        network = plugin.get_network(dbcontext, port['network_id'])
        port[addr_pair.ADDRESS_PAIRS] = (
            plugin.get_allowed_address_pairs(dbcontext, port['id']))
        plugin.extend_port_extra_dhcp_opts_dict(dbcontext, port)
        cls.filter_create_attributes(port, dbcontext, network)
        return plugin.get_bound_port_context(dbcontext, port['id'])

    @classmethod
    def filter_create_attributes_without_plugin(cls, port, dbcontext, network):
        binding = {}
        port[addr_pair.ADDRESS_PAIRS] = []
        context = ml2_context.PortContext(
            None, dbcontext, port, network, binding, None)
        cls.filter_create_attributes(port, context, network)
        return context


class SecurityGroupFilter(ResourceFilterBase):
    @classmethod
    def filter_create_attributes_with_plugin(cls, sg, plugin, dbcontext):
        context = driver_context.SecurityGroupContext(plugin, dbcontext, sg)
        return context

    @classmethod
    def filter_create_attributes_without_plugin(cls, sg, dbcontext, network):
        return cls.filter_create_attributes_with_plugin(sg, None, dbcontext)


class SecurityGroupRuleFilter(ResourceFilterBase):
    @classmethod
    def filter_create_attributes_with_plugin(cls, sg_rule, plugin, dbcontext):
        context = driver_context.SecurityGroupRuleContext(
            plugin, dbcontext, sg_rule)
        return context

    @classmethod
    def filter_create_attributes_without_plugin(
            cls, sg_rule, dbcontext, network):
        return cls.filter_create_attributes_with_plugin(
            sg_rule, None, dbcontext)



# Copyright (c) 2013 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import six

import l3_agent_scheduler

from eventlet import greenthread
from neutron.db import agents_db
from neutron.db import agentschedulers_db
from neutron.db import l3_agentschedulers_db
from neutron.db import l3_db
from neutron.db import l3_hamode_db
from neutron.db import models_v2
from neutron.extensions import agent as ext_agent
from neutron.extensions import l3_ext_ha_mode as l3_ha
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from sqlalchemy.orm import exc as sa_exc

__author__ = '10069907'

LOG = logging.getLogger(__name__)
cfg.CONF.register_opts(l3_hamode_db.L3_HA_OPTS)


def method_patch(cls, bases, dct):
    base = bases[0]
    for name, value in dct.iteritems():
        if name not in("__module__", "__metaclass__", "__doc__"):
            setattr(base, name, value)
    return base


@six.add_metaclass(method_patch)
class AddAgentToRouterBinding(agents_db.AgentDbMixin):
    def _get_all_routers_with_interface_or_gateway(self, context):
        try:
            port_db = (context.session.query(models_v2.Port).
                       enable_eagerloads(False).
                       filter(models_v2.Port.device_owner.
                              startswith("network:router_interface")).
                       all())
        except sa_exc.NoResultFound:
            LOG.debug("No ports have port_id starting with router_interface")
            return None

        try:
            router_db = (context.session.query(l3_db.Router).
                         enable_eagerloads(False).all())
        except sa_exc.NoResultFound:
            LOG.debug("No router have Found")
            return None
        router_ids = [router.id for router in router_db]
        return [
            port.device_id for port in port_db if port.device_id in router_ids
            ]

    def _binding_router_to_agent(self, context, router_ids, agent_id):
        for r_id in router_ids:
            LOG.debug('_binding_router_to_agent, process:%s-%s' %
                      (r_id, agent_id))
            query = context.session.query(l3_agentschedulers_db.
                                          RouterL3AgentBinding)
            query = query.filter(l3_agentschedulers_db.RouterL3AgentBinding.
                                 router_id == r_id,
                                 l3_agentschedulers_db.RouterL3AgentBinding.
                                 l3_agent_id == agent_id)
            try:
                query.one()
                LOG.debug('find old router_agent_binding %s-%s '
                          % (r_id, agent_id))
            except sa_exc.NoResultFound:
                with context.session.begin(subtransactions=True):
                    binding = l3_agentschedulers_db.RouterL3AgentBinding()
                    binding.l3_agent_id = agent_id
                    binding.router_id = r_id
                    try:
                        context.session.add(binding)
                        LOG.debug('add router_agent_binding %s-%s'
                                  % (r_id, agent_id))
                    except db_exc.DBError:
                        LOG.debug('add router_agent_binding %s-%s fail'
                                  % (r_id, agent_id))

    def _update_l3_agent_router_bindings(self, context, agent_id):
        router_ids_need_binding = \
            self._get_all_routers_with_interface_or_gateway(context)
        LOG.info('get all_router_with_port:%(r_id)s,and agent:%(agt_id)s',
                 {'r_id': router_ids_need_binding, 'agt_id': agent_id})
        if router_ids_need_binding is None:
            return
        self._binding_router_to_agent(context,
                                      router_ids_need_binding, agent_id)

    def _create_or_update_agent(self, context, agent):
        with context.session.begin(subtransactions=True):
            res_keys = ['agent_type', 'binary', 'host', 'topic']
            res = dict((k, agent[k]) for k in res_keys)

            configurations_dict = agent.get('configurations', {})
            res['configurations'] = jsonutils.dumps(configurations_dict)
            res['load'] = self._get_agent_load(agent)
            current_time = timeutils.utcnow()
            try:
                agent_db = self._get_agent_by_type_and_host(
                    context, agent['agent_type'], agent['host'])
                if agent['topic'] == "l3_agent":
                    self._update_l3_agent_router_bindings(context, agent_db.id)
                res['heartbeat_timestamp'] = current_time
                if agent.get('start_flag'):
                    res['started_at'] = current_time
                greenthread.sleep(0)
                agent_db.update(res)
            except ext_agent.AgentNotFoundByTypeHost:
                greenthread.sleep(0)
                res['created_at'] = current_time
                res['started_at'] = current_time
                res['heartbeat_timestamp'] = current_time
                res['admin_state_up'] = True
                agent_db = agents_db.Agent(**res)
                greenthread.sleep(0)
                context.session.add(agent_db)
            greenthread.sleep(0)


class AllL3AgentsScheduler(l3_agent_scheduler.L3Scheduler):
    """allocate all L3 agent for a router."""

    def get_all_l3_agents(self, plugin, context):
        """Return L3 agents where a router could be scheduled."""
        with context.session.begin(subtransactions=True):
            query = context.session.query(agents_db.Agent)
            query = query.filter(
                agents_db.Agent.topic == 'l3_agent')
            query = (query.filter_by(admin_state_up=True))

            return [l3_agent
                    for l3_agent in query
                    if (agentschedulers_db.AgentSchedulerDbMixin.
                        is_eligible_agent(True, l3_agent))]

    def _schedule_router(self, plugin, context, router_id, candidates=None):
        sync_router = plugin.get_router(context, router_id)
        router_distributed = sync_router.get('distributed', False)
        if router_distributed or sync_router.get('ha', False):
            raise l3_ha.DistributedHARouterNotSupported()
        if candidates is None:
            candidates = self.get_all_l3_agents(plugin, context)
            LOG.debug('all_cap:%s' % candidates)
        if not candidates:
            return
        for chosen_agent in candidates:
            query = context.session.query(
                l3_agentschedulers_db.RouterL3AgentBinding)
            query = query.filter(
                l3_agentschedulers_db.RouterL3AgentBinding.router_id ==
                router_id,
                l3_agentschedulers_db.RouterL3AgentBinding.l3_agent_id ==
                chosen_agent['id'])
            try:
                query.one()
                LOG.debug('find old router_agent_binding %s-%s '
                          % (router_id, chosen_agent['id']))
            except sa_exc.NoResultFound:
                self.bind_router(context, router_id, chosen_agent)
        return chosen_agent

    def schedule(self, plugin, context, router_id, candidates=None):
        LOG.debug('AllL3AgentsScheduler.schedule')
        return self._schedule_router(
            plugin, context, router_id, candidates=candidates)

    def _choose_router_agent(self, plugin, context, candidates):
        """Choose an agent from candidates based on a specific policy."""
        pass

    def _choose_router_agents_for_ha(self, plugin, context, candidates):
        num_agents = self.get_num_of_agents_for_ha(len(candidates))
        return random.sample(candidates, num_agents)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
This module manages the HTTP and HTTPS connections to the backend controllers.

The main class it provides for external use is ServerPool which manages a set
of ServerProxy objects that correspond to individual backend controllers.

The following functionality is handled by this module:
- Translation of rest_* function calls to HTTP/HTTPS calls to the controllers
- Automatic failover between controllers
- SSL Certificate enforcement
- HTTP Authentication

"""
import base64
import eventlet
import httplib
import json
import os
import re
import socket
import ssl

from oslo_log import log
from oslo_serialization import jsonutils
try:
    from oslo.config import cfg
except Exception:
    from oslo_config import cfg
try:
    from neutron._i18n import _
    from neutron_lib import exceptions
except Exception:
    from neutron.common import exceptions

LOG = log.getLogger(__name__)

TOPOLOGY_PATH = "/topology"
SUCCESS_CODES = range(200, 207)
FAILURE_CODES = [301, 302, 303]
BASE_URI = '/'
response_code = {
    'OK': 200,
    'OK_1': 204,
    'Bad_Request': 400,
    'Unauthorized': 401,
    'Forbidden': 403,
    'Not_Found': 404,
    'Conflict': 409,
}


class RemoteRestError(exceptions.NeutronException):
    message = _("Error in REST call to remote network "
                "controller: %(reason)s")
    status = None

    def __init__(self, **kwargs):
        self.status = kwargs.pop('status', None)
        self.reason = kwargs.get('reason')
        super(RemoteRestError, self).__init__(**kwargs)


class ServerProxy(object):
    """REST server proxy to a network controller."""

    def __init__(self, server, port, ssl, auth, timeout,
                 base_uri, success_codes, name, combined_cert):
        self.server = server
        self.port = port
        self.ssl = ssl
        self.base_uri = base_uri
        self.timeout = timeout
        self.name = name
        self.success_codes = success_codes
        self.auth = auth
        self.failed = False
        self.capabilities = []
        self.currentconn = None
        self.header = {
            "Accept": "application/json",
            "Content-Type": "application/json"}
        self.combined_cert = combined_cert
        self.auth_type = self.get_auth_type()

        self.username, self.password = re.split(':', self.auth)

        if self.auth_type == "BASE64":
            self.auth = "Basic " + \
                base64.b64encode("%s:%s" % (self.username, self.password))
        elif self.auth_type == "AES":
            encrypt_pwd = self.add_aes_encryption(self.password)
            self.auth = "Basic " + \
                base64.b64encode("%s:%s" % (self.username, encrypt_pwd))
            LOG.debug('after base64enc auth=%s' % "****")

        if self.auth is not None:
            self.header["Authorization"] = self.auth
            self.header["Realm"] = 'ZENIC'

    def get_auth_type(self):
        url = '/rest/v1/security/encryption'
        method = "GET"
        body = None
        r = self.send_http_request(method, url, body)
        if r:
            jsn = json.loads(r)
            enc = jsn['content']
            return enc
        else:
            return ""

    def send_http_request(self, method="", url="", body=(), trans=False):
        if self.ssl and not self.combined_cert:
            http = HTTPSConnectionWithValidation(self.server, self.port)
        else:
            http = httplib.HTTPConnection(self.server, self.port)
        if trans:
            payload = body
        else:
            if body:
                payload = json.dumps(body)
            else:
                payload = None

        LOG.debug('servermanager send_http_request, method=%s,header=%s, '
                  'url=%s,payload=%s' % (method, self.header, url, "****"))
        http.request(method, url, payload, self.header)
        res = http.getresponse()
        data = res.read()
        LOG.info(_(
            'servermanager send_http_request --ret code=%s ---'),
            res.status)
        if (res.status not in (response_code['OK'], response_code['OK_1'])):
            if res.status == 404:
                LOG.debug(
                    'servermanager send_http_request finished, '
                    'res code=%s, but get none data' % res.status)
            else:
                LOG.error(_(
                    'servermanager send_http_request failed, res code=%('
                    'code)s, reason=%(data)s'), {'code': res.status,
                                                 'data': data})
            return None
        LOG.info(_('servermanager send_http_request --ret data=%s ---'),
                 "****")

        return data

    def add_aes_encryption(self, pwd=""):
        url = '/rest/v1/security/encrypt/AES'
        method = "POST"
        body = 'pwd=%s' % pwd
        r = self.send_http_request(method, url, body, True)
        return r

    def rest_call(self, action, resource, data='', headers={}, timeout=False,
                  reconnect=False):
        uri = self.base_uri + resource
        body = jsonutils.dumps(data)
        if not headers:
            headers = {}
        headers['Content-type'] = 'application/json'
        headers['Accept'] = 'application/json'
        headers['NeutronProxy-Agent'] = self.name

        if self.auth:
            headers['Authorization'] = self.auth
            headers['Realm'] = 'ZENIC'

        LOG.info(("ServerProxy: server=%(server)s, port=%(port)d, "
                  "ssl=%(ssl)r"),
                 {'server': self.server, 'port': self.port, 'ssl': self.ssl})
        LOG.info(("ServerProxy: resource=%(resource)s, data=%(data)r, "
                  "action=%(action)s"),
                 {'resource': resource, 'data': data,
                  'action': action})

        conn = None
        if self.ssl and not self.combined_cert:
            conn = HTTPSConnectionWithValidation(
                    self.server, self.port, timeout=self.timeout)
            if conn is None:
                LOG.error((
                    'ServerProxy: Could not establish HTTPS connection'))
                return 0, None, None, None
        else:
            conn = httplib.HTTPConnection(
                self.server, self.port, timeout=self.timeout)
            if conn is None:
                LOG.error((
                    'ServerProxy: Could not establish HTTP connection'))
                return 0, None, None, None

        try:
            conn.request(action, uri, body, headers)
            response = conn.getresponse()
            respstr = response.read()
            respdata = respstr
            if response.status in self.success_codes:
                try:
                    respdata = jsonutils.loads(respstr)
                except ValueError:
                    # response was not JSON, ignore the exception
                    pass
            ret = (response.status, response.reason, respstr, respdata)
        except socket.timeout as e1:
            LOG.error(('ServerProxy: %(action)s failure, %(el)r'),
                      {"action": action, "el": e1})
            ret = 9, None, None, None
        except socket.error as e:
            LOG.error(("ServerProxy: %(action)s failure, %(e)r"),
                      {"action": action, "e": e})
            ret = 0, None, None, None
        conn.close()
        LOG.info(_("ServerProxy: status=%(status)d, reason=%(reason)r, "
                   "ret=%(ret)s, data=%(data)r"), {'status': ret[0],
                                                   'reason': ret[1],
                                                   'ret': ret[2],
                                                   'data': ret[3]})
        return ret

        """
        if not self.currentconn or reconnect:
            if self.currentconn:
                self.currentconn.close()
            if self.ssl:
                self.currentconn = HTTPSConnectionWithValidation(
                    self.server, self.port, timeout=timeout)
                if self.currentconn is None:
                    LOG.error(_('ServerProxy: Could not establish HTTPS '
                                'connection'))
                    return 0, None, None, None
                self.currentconn.combined_cert = self.combined_cert
            else:
                self.currentconn = httplib.HTTPConnection(
                    self.server, self.port, timeout=timeout)
                if self.currentconn is None:
                    LOG.error(_('ServerProxy: Could not establish HTTP '
                                'connection'))
                    return 0, None, None, None

        try:
            self.currentconn.request(action, uri, body, headers)
            response = self.currentconn.getresponse()
            respstr = response.read()
            respdata = respstr
            if response.status in self.success_codes:
                try:
                    respdata = json.loads(respstr)
                except ValueError:
                    # response was not JSON, ignore the exception
                    pass
            ret = (response.status, response.reason, respstr, respdata)
        except httplib.HTTPException:
            # If we were using a cached connection, try again with a new one.
            with excutils.save_and_reraise_exception() as ctxt:
                self.currentconn.close()
                if reconnect:
                    # if reconnect is true, this was on a fresh connection so
                    # reraise since this server seems to be broken
                    ctxt.reraise = True
                else:
                    # if reconnect is false, it was a cached connection so
                    # try one more time before re-raising
                    ctxt.reraise = False
            return self.rest_call(action, resource, data, headers,
                                  timeout=timeout, reconnect=True)
        except (socket.timeout, socket.error) as e:
            self.currentconn.close()
            LOG.error(_('ServerProxy: %(action)s failure, %(e)r'),
                      {'action': action, 'e': e})
            ret = 0, None, None, None
        LOG.debug(_("ServerProxy: status=%(status)d, reason=%(reason)r, "
                    "ret=%(ret)s, data=%(data)r"), {'status': ret[0],
                                                    'reason': ret[1],
                                                    'ret': ret[2],
                                                    'data': ret[3]})
        return ret
        """


class ServerPool(object):

    def __init__(self, servers, auth, ssl, no_ssl_validation, ssl_sticky,
                 ssl_cert_directory, consistency_interval,
                 timeout=False, cache_connections=False,
                 base_uri=BASE_URI, success_codes=SUCCESS_CODES,
                 failure_codes=FAILURE_CODES, name='NeutronRestProxy'):
        LOG.debug(_("ServerPool: initializing"))
        # 'servers' is the list of network controller REST end-points
        # (used in order specified till one succeeds, and it is sticky
        # till next failure). Use 'server_auth' to encode api-key
        self.auth = auth
        self.ssl = ssl
        self.base_uri = base_uri
        self.success_codes = success_codes
        self.failure_codes = failure_codes
        self.name = name
        self.timeout = timeout
        self.always_reconnect = not cache_connections
        self.consistency_interval = consistency_interval
        self.no_ssl_validation = no_ssl_validation
        self.ssl_cert_directory = ssl_cert_directory
        self.ssl_sticky = ssl_sticky

        default_port = 8000
        if timeout is not False:
            self.timeout = timeout

        # Function to use to retrieve topology for consistency syncs.
        # Needs to be set by module that uses the servermanager.
        self.get_topo_function = None
        self.get_topo_function_args = {}

        if not servers:
            raise cfg.Error(_('Servers not defined. Aborting server manager.'))
        servers = [s if len(s.rsplit(':', 1)) == 2
                   else "%s:%d" % (s, default_port)
                   for s in servers]
        if any((len(spl) != 2 or not spl[1].isdigit())
               for spl in [sp.rsplit(':', 1)
                           for sp in servers]):
            raise cfg.Error(_('Servers must be defined as <ip>:<port>. '
                              'Configuration was %s') % servers)
        self.servers = [
            self.server_proxy_for(server, int(port))
            for server, port in (s.rsplit(':', 1) for s in servers)
        ]
        eventlet.spawn(self._consistency_watchdog, self.consistency_interval)
        LOG.debug("ServerPool: initialization done")

    def server_proxy_for(self, server, port):
        combined_cert = self._get_combined_cert_for_server(server, port)
        return ServerProxy(server, port, self.ssl, self.auth,
                           self.timeout, self.base_uri, self.success_codes,
                           self.name, combined_cert=combined_cert)

    def _get_combined_cert_for_server(self, server, port):
        # The ssl library requires a combined file with all trusted certs
        # so we make one containing the trusted CAs and the corresponding
        # host cert for this server
        combined_cert = None
        if self.ssl and not self.no_ssl_validation:
            base_ssl = self.ssl_cert_directory
            host_dir = os.path.join(base_ssl, 'host_certs')
            ca_dir = os.path.join(base_ssl, 'ca_certs')
            combined_dir = os.path.join(base_ssl, 'combined')
            combined_cert = os.path.join(combined_dir, '%s.pem' % server)
            if not os.path.exists(base_ssl):
                raise cfg.Error(_('ssl_cert_directory [%s] does not exist. '
                                  'Create it or disable ssl.') % base_ssl)
            for automake in [combined_dir, ca_dir, host_dir]:
                if not os.path.exists(automake):
                    os.makedirs(automake)

            # get all CA certs
            certs = self._get_ca_cert_paths(ca_dir)

            # check for a host specific cert
            hcert, exists = self._get_host_cert_path(host_dir, server)
            if exists:
                certs.append(hcert)
            elif self.ssl_sticky:
                self._fetch_and_store_cert(server, port, hcert)
                certs.append(hcert)
            if not certs:
                raise cfg.Error(_('No certificates were found to verify '
                                  'controller %s') % (server))
            self._combine_certs_to_file(certs, combined_cert)
        return combined_cert

    def _combine_certs_to_file(self, certs, cfile):
        """
        Concatenates the contents of each certificate in a list of
        certificate paths to one combined location for use with ssl
        sockets.
        """
        with open(cfile, 'w') as combined:
            for c in certs:
                with open(c, 'r') as cert_handle:
                    combined.write(cert_handle.read())

    def _get_host_cert_path(self, host_dir, server):
        """
        returns full path and boolean indicating existence
        """
        hcert = os.path.join(host_dir, '%s.pem' % server)
        if os.path.exists(hcert):
            return hcert, True
        return hcert, False

    def _get_ca_cert_paths(self, ca_dir):
        certs = [os.path.join(root, name)
                 for name in [
                     name for (root, dirs, files) in os.walk(ca_dir)
                     for name in files
        ]
            if name.endswith('.pem')]
        return certs

    def _fetch_and_store_cert(self, server, port, path):
        """
        Grabs a certificate from a server and writes it to
        a given path.
        """
        try:
            cert = ssl.get_server_certificate((server, port))
        except Exception as e:
            raise cfg.Error(_('Could not retrieve initial '
                              'certificate from controller %(server)s. '
                              'Error details: %(error)s') %
                            {'server': server, 'error': str(e)})

        LOG.warning("Storing to certificate for host %(server)s "
                    "at %(path)s" % {'server': server, 'path': path})
        self._file_put_contents(path, cert)

        return cert

    def _file_put_contents(self, path, contents):
        # Simple method to write to file.
        # Created for easy Mocking
        with open(path, 'w') as handle:
            handle.write(contents)

    def server_failure(self, resp, ignore_codes=[]):
        """Define failure codes as required.

        Note: We assume 301-303 is a failure, and try the next server in
        the server pool.
        """
        return resp[0] in self.failure_codes and resp[0] not in ignore_codes

    def action_success(self, resp):
        """Defining success codes as required.

        Note: We assume any valid 2xx as being successful response.
        """
        return resp[0] in self.success_codes

    def rest_call(self, action, resource, data, headers, ignore_codes,
                  timeout=False):
        good_first = sorted(self.servers, key=lambda x: x.failed)
        first_response = None
        for active_server in good_first:
            ret = active_server.rest_call(action, resource, data, headers,
                                          timeout,
                                          reconnect=self.always_reconnect)
            # If inconsistent, do a full synchronization
            # if ret[0] == httplib.CONFLICT:
            #    if self.get_topo_function:
            #        data = self.get_topo_function(
            #            **self.get_topo_function_args)
            #        active_server.rest_call('PUT', TOPOLOGY_PATH, data,
            #                                timeout=None)
            # Store the first response as the error to be bubbled up to the
            # user since it was a good server. Subsequent servers will most
            # likely be cluster slaves and won't have a useful error for the
            # user (e.g. 302 redirect to master)
            if not first_response:
                first_response = ret
            if not self.server_failure(ret, ignore_codes):
                active_server.failed = False
                return ret
            else:
                try:
                    LOG.error('ServerProxy: %(action)s failure for '
                              'servers:%(server)r Response: %(response)s',
                              {'action': action,
                               'server': (active_server.server,
                                          active_server.port),
                               'response': unicode(ret[3], "utf-8")})
                    LOG.error("ServerProxy: Error details: status=%(status)d,"
                              " reason=%(reason)r, ret=%(ret)s, data=%(data)r",
                              {'status': ret[0], 'reason': ret[1],
                               'ret': unicode(ret[2], "utf-8"),
                               'data': unicode(ret[3], "utf-8")})
                except Exception as e:
                    LOG.error("fail to display info, err: %(e)s",
                              {'e': e})
                active_server.failed = True

        # All servers failed, reset server list and try again next time
        LOG.error(_('ServerProxy: %(action)s failure for all servers: '
                    '%(server)r'),
                  {'action': action,
                   'server': tuple((s.server,
                                    s.port) for s in self.servers)})
        return first_response

    def rest_action(self, action, resource, data='', errstr='%s',
                    ignore_codes=[], headers={}, timeout=False):
        """
        Wrapper for rest_call that verifies success and raises a
        RemoteRestError on failure with a provided error string
        By default, 404 errors on DELETE calls are ignored because
        they already do not exist on the backend.
        """
        LOG.debug(_("rest_action: %(action)s action to "
                    "resource %(resource)s %(data)s"),
                  {'action': action, 'resource': resource, 'data': data})

        resp = self.rest_call(action, resource, data, headers, ignore_codes,
                              timeout)
        if self.server_failure(resp, ignore_codes):
            try:
                LOG.error(errstr, unicode(resp[2], "utf-8"))  # noqa
            except Exception as e:
                LOG.error("fail to display info, err: %(e)s",
                          {'e': e})
            raise RemoteRestError(reason=resp[2], status=resp[0])
        if resp[0] in ignore_codes:
            LOG.warning(_("NeutronRestProxyV2: Received and ignored error "
                          "code %(code)s on %(action)s action to resource "
                          "%(resource)s"),
                        {'code': resp[2], 'action': action,
                         'resource': resource})
        return resp

    def _consistency_watchdog(self, polling_interval=60):
        return


class HTTPSConnectionWithValidation(httplib.HTTPSConnection):

    # If combined_cert is None, the connection will continue without
    # any certificate validation.
    combined_cert = None

    def connect(self):
        sock = socket.create_connection((self.host, self.port),
                                        self.timeout, self.source_address)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        if self.combined_cert:
            self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file,
                                        cert_reqs=ssl.CERT_REQUIRED,
                                        ca_certs=self.combined_cert)
        else:
            self.sock = ssl.wrap_socket(sock, self.key_file,
                                        self.cert_file,
                                        cert_reqs=ssl.CERT_NONE)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_zte.fwaas.zenic_firewall_pool import ZenicRestServerPool
try:
    from neutron_lib import exceptions
except Exception:
    from neutron.common import exceptions
from neutron.plugins.common import constants as const
from neutron_fwaas.services.firewall.fwaas_plugin import FirewallPlugin
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

rest_proxy_opts = [
    cfg.ListOpt('servers', default=['localhost:8800'],
                help=_("A comma separated list of Big Switch or Floodlight "
                       "servers and port numbers. The plugin proxies the "
                       "requests to the Big Switch/Floodlight server, "
                       "which performs the networking configuration. Only one"
                       "server is needed per deployment, but you may wish to"
                       "deploy multiple servers to support failover.")),
    cfg.StrOpt('server_auth', default=None, secret=True,
               help=_("The username and password for authenticating against "
                      " the Big Switch or Floodlight controller.")),
    cfg.StrOpt('base_url', default='/restconf/operations/zenic-fwapp-model:',
               help=_("Base URL for restconf server ")),
    cfg.BoolOpt('server_ssl', default=False,
                help=_("If True, Use SSL when connecting to the Big Switch or "
                       "Floodlight controller.")),
    cfg.BoolOpt('ssl_sticky', default=True,
                help=_("Trust and store the first certificate received for "
                       "each controller address and use it to validate future "
                       "connections to that address.")),
    cfg.BoolOpt('no_ssl_validation', default=False,
                help=_("Disables SSL certificate validation for controllers")),
    cfg.BoolOpt('cache_connections', default=True,
                help=_("Re-use HTTP/HTTPS connections to the controller.")),
    cfg.StrOpt('ssl_cert_directory',
               default='/etc/neutron/plugins/proxyagent/znic/ssl',
               help=_("Directory containing ca_certs and host_certs "
                      "certificate directories.")),
    cfg.BoolOpt('sync_data', default=False,
                help=_("Sync data on connect")),
    cfg.BoolOpt('auto_sync_on_failure', default=True,
                help=_("If neutron fails to create a resource because "
                       "the backend controller doesn't know of a dependency, "
                       "the plugin automatically triggers a full data "
                       "synchronization to the controller.")),
    cfg.IntOpt('consistency_interval', default=60,
               help=_("Time between verifications that the backend controller "
                      "database is consistent with Neutron")),
    cfg.IntOpt('server_timeout', default=10,
               help=_("Maximum number of seconds to wait for proxy request "
                      "to connect and complete.")),
    cfg.IntOpt('thread_pool_size', default=4,
               help=_("Maximum number of threads to spawn to handle large "
                      "volumes of port creations.")),
]


def get_rest_proxy_conf():
    conf_proxy = cfg.ConfigOpts()
    conf_proxy.register_opts(rest_proxy_opts, 'RESTPROXY')
    conf_proxy(args=[], default_config_files=['/etc/neutron/plugin.ini'])
    return conf_proxy


class PostZenicRestError(exceptions.NeutronException):
    message = _("The error occurred when posting to zenic: %(reason)s")

    def __init__(self, **kwargs):
        self.msg = self.message % kwargs


class ZenicFirewallPlugin(FirewallPlugin):

    """Implementation of the Zenic Firewall Service Plugin.

    This class manages the workflow of FWaaS request/response.
    Most DB related works are implemented in class
    firewall_db.Firewall_db_mixin.
    """

    def __init__(self):
        conf_proxy = get_rest_proxy_conf()
        if conf_proxy.RESTPROXY.servers is not '':
            self.servers = ZenicRestServerPool(
                conf_proxy.RESTPROXY.servers,
                conf_proxy.RESTPROXY.server_auth,
                conf_proxy.RESTPROXY.server_ssl,
                conf_proxy.RESTPROXY.no_ssl_validation,
                conf_proxy.RESTPROXY.ssl_sticky,
                conf_proxy.RESTPROXY.ssl_cert_directory,
                conf_proxy.RESTPROXY.consistency_interval,
                conf_proxy.RESTPROXY.server_timeout,
                conf_proxy.RESTPROXY.cache_connections,
                conf_proxy.RESTPROXY.base_url)
        super(ZenicFirewallPlugin, self).__init__()

    def create_firewall(self, context, firewall):
        LOG.info(_("create_firewall firewall=%s"), firewall)
        fw = super(ZenicFirewallPlugin, self).create_firewall(
            context, firewall)
        LOG.info(_("create_firewall firewall=%s"), fw)
        try:
            self.servers.create_firewall(context, fw)
            self.set_firewall_status(context, fw['id'], const.ACTIVE)
        except BaseException:
            super(ZenicFirewallPlugin, self).delete_firewall(context,
                                                             fw['id'])
            raise PostZenicRestError(reason="create firewall!")
        return fw

    def update_firewall(self, context, id, firewall):
        LOG.info(_("update_firewall firewall=%s"), firewall)
        try:
            self.servers.update_firewall(context, id, firewall)
        except BaseException:
            raise PostZenicRestError(reason="update firewall!")
        fw = super(ZenicFirewallPlugin, self).update_firewall(
            context, id, firewall)
        self.set_firewall_status(context, id, const.ACTIVE)
        return fw

    def delete_firewall(self, context, id):
        LOG.info(_("delete_firewall id = %s"), id)
        try:
            self.servers.delete_firewall(context, id)
        except BaseException:
            raise PostZenicRestError(reason="delete firewall!")
        super(ZenicFirewallPlugin, self).delete_firewall(context, id)
        self.firewall_deleted(context, id)

    def create_firewall_policy(self, context, firewall_policy):
        LOG.info(_("create_firewall_policy firewall_policy=%s"),
                 firewall_policy)
        fwp = super(ZenicFirewallPlugin,
                    self).create_firewall_policy(context, firewall_policy)
        LOG.info(_("create_firewall_policy firewall_policy=%s"), fwp)
        try:
            self.servers.create_firewall_policy(context, fwp)
        except BaseException:
            super(ZenicFirewallPlugin,
                  self).delete_firewall_policy(context, fwp['id'])
            raise PostZenicRestError(reason="create firewall policy!")
        return fwp

    def update_firewall_policy(self, context, id, firewall_policy):
        LOG.info(_("update_firewall_policy firewall_policy=%s"),
                 firewall_policy)
        try:
            self.servers.update_firewall_policy(context, id, firewall_policy)
        except BaseException:
            raise PostZenicRestError(reason="update firewall policy!")
        fwp = super(ZenicFirewallPlugin,
                    self).update_firewall_policy(context, id, firewall_policy)
        self.update_policy_related_firewall_status(context, id, const.ACTIVE)
        return fwp

    def delete_firewall_policy(self, context, id):
        LOG.info(_("delete_firewall_policy id = %s"), id)
        try:
            self.servers.delete_firewall_policy(context, id)
        except BaseException:
            raise PostZenicRestError(reason="delete firewall policy!")
        super(ZenicFirewallPlugin, self).delete_firewall_policy(context, id)

    def create_firewall_rule(self, context, firewall_rule):
        LOG.info(_("create_firewall_policy firewall_rule=%s"), firewall_rule)
        fwr = super(ZenicFirewallPlugin,
                    self).create_firewall_rule(context, firewall_rule)
        LOG.info(_("create_firewall_policy firewall_rule=%s"), fwr)
        try:
            self.servers.create_firewall_rule(context, fwr)
        except BaseException:
            super(ZenicFirewallPlugin,
                  self).delete_firewall_rule(context, fwr['id'])
            raise PostZenicRestError(reason="create firewall rule!")
        return fwr

    def update_firewall_rule(self, context, id, firewall_rule):
        LOG.info(_("upate_firewall_rule firewall_rule=%s"), firewall_rule)
        try:
            self.servers.update_firewall_rule(context, id, firewall_rule)
        except BaseException:
            raise PostZenicRestError(reason="update firewall rule!")
        fwr = super(ZenicFirewallPlugin,
                    self).update_firewall_rule(context, id, firewall_rule)
        firewall_policy_id = fwr['firewall_policy_id']
        if firewall_policy_id:
            self.update_policy_related_firewall_status(
                context, firewall_policy_id, const.ACTIVE)
        return fwr

    def delete_firewall_rule(self, context, id):
        LOG.info(_("delete_firewall_rule id = %s"), id)
        try:
            self.servers.delete_firewall_rule(context, id)
        except BaseException:
            raise PostZenicRestError(reason="delete firewall policy!")
        super(ZenicFirewallPlugin, self).delete_firewall_rule(context, id)

    def insert_rule(self, context, id, rule_info):
        LOG.info(_("insert_rule rule_info=%s"), rule_info)
        try:
            self.servers.insert_rule(context, id, rule_info)
        except BaseException:
            raise PostZenicRestError(reason="insert rule!")
        fwp = super(ZenicFirewallPlugin,
                    self).insert_rule(context, id, rule_info)
        self.update_policy_related_firewall_status(context, id, const.ACTIVE)
        return fwp

    def remove_rule(self, context, id, rule_info):
        LOG.info(_("remove_rule rule_info=%s"), rule_info)
        try:
            self.servers.remove_rule(context, id, rule_info)
        except BaseException:
            raise PostZenicRestError(reason="remove rule!")
        fwp = super(ZenicFirewallPlugin,
                    self).remove_rule(context, id, rule_info)
        self.update_policy_related_firewall_status(context, id, const.ACTIVE)
        return fwp

    def set_firewall_status(self, context, firewall_id, status):
        """uses this to set a firewall's status."""
        LOG.debug("set_firewall_status() called")
        with context.session.begin(subtransactions=True):
            fw_db = self._get_firewall(context, firewall_id)
            # ignore changing status if firewall expects to be deleted
            # That case means that while some pending operation has been
            # performed on the backend, neutron server received delete request
            # and changed firewall status to const.PENDING_DELETE
            if fw_db.status == const.PENDING_DELETE:
                LOG.debug("Firewall %(fw_id)s in PENDING_DELETE state, "
                          "not changing to %(status)s",
                          {'fw_id': firewall_id, 'status': status})
                return False
            if status in (const.ACTIVE, const.DOWN, const.INACTIVE):
                fw_db.status = status
                return True
            else:
                fw_db.status = const.ERROR
                return False

    def firewall_deleted(self, context, firewall_id):
        """Agent uses this to indicate firewall is deleted."""
        LOG.debug("firewall_deleted() called")
        with context.session.begin(subtransactions=True):
            fw_db = self._get_firewall(context, firewall_id)
            # allow to delete firewalls in ERROR state
            if fw_db.status in (const.PENDING_DELETE, const.ERROR):
                self.delete_db_firewall_object(context, firewall_id)
                return True
            else:
                LOG.warn(_('Firewall %(fw)s unexpectedly deleted by agent, '
                           'status was %(status)s'),
                         {'fw': firewall_id, 'status': fw_db.status})
                fw_db.status = const.ERROR
                return False

    def update_policy_related_firewall_status(
            self, context, firewall_policy_id, status):
        firewall_policy = self.get_firewall_policy(context, firewall_policy_id)
        if firewall_policy:
            for firewall_id in firewall_policy['firewall_list']:
                self.set_firewall_status(context, firewall_id, status)



# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_zte.qos.zenic_qos_pool import ZenicRestServerPool
try:
    from neutron_lib import exceptions
except Exception:
    from neutron.common import exceptions
try:
    from neutron.services.qos.qos_plugin import QoSPlugin
except Exception:
    pass
try:
    from neutron.objects.qos import rule as rule_object
except Exception:
    pass
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

rest_proxy_opts = [
    cfg.ListOpt('servers', default=['localhost:8800'],
                help=_("A comma separated list of Big Switch or Floodlight "
                       "servers and port numbers. The plugin proxies the "
                       "requests to the Big Switch/Floodlight server, "
                       "which performs the networking configuration. Only one"
                       "server is needed per deployment, but you may wish to"
                       "deploy multiple servers to support failover.")),
    cfg.StrOpt('server_auth', default=None, secret=True,
               help=_("The username and password for authenticating against "
                      " the Big Switch or Floodlight controller.")),
    cfg.StrOpt('base_url', default='/restconf/operations/zenic-vdcapp-model:',
               help=_("Base URL for restconf server ")),
    cfg.BoolOpt('server_ssl', default=False,
                help=_("If True, Use SSL when connecting to the Big Switch or "
                       "Floodlight controller.")),
    cfg.BoolOpt('ssl_sticky', default=True,
                help=_("Trust and store the first certificate received for "
                       "each controller address and use it to validate future "
                       "connections to that address.")),
    cfg.BoolOpt('no_ssl_validation', default=False,
                help=_("Disables SSL certificate validation for controllers")),
    cfg.BoolOpt('cache_connections', default=True,
                help=_("Re-use HTTP/HTTPS connections to the controller.")),
    cfg.StrOpt('ssl_cert_directory',
               default='/etc/neutron/plugins/proxyagent/znic/ssl',
               help=_("Directory containing ca_certs and host_certs "
                      "certificate directories.")),
    cfg.BoolOpt('sync_data', default=False,
                help=_("Sync data on connect")),
    cfg.BoolOpt('auto_sync_on_failure', default=True,
                help=_("If neutron fails to create a resource because "
                       "the backend controller doesn't know of a dependency, "
                       "the plugin automatically triggers a full data "
                       "synchronization to the controller.")),
    cfg.IntOpt('consistency_interval', default=60,
               help=_("Time between verifications that the backend controller "
                      "database is consistent with Neutron")),
    cfg.IntOpt('server_timeout', default=10,
               help=_("Maximum number of seconds to wait for proxy request "
                      "to connect and complete.")),
    cfg.IntOpt('thread_pool_size', default=4,
               help=_("Maximum number of threads to spawn to handle large "
                      "volumes of port creations.")),
]


def get_rest_proxy_conf():
    conf_proxy = cfg.ConfigOpts()
    conf_proxy.register_opts(rest_proxy_opts, 'RESTPROXY')
    conf_proxy(args=[], default_config_files=['/etc/neutron/plugin.ini'])
    return conf_proxy


class PostZenicRestError(exceptions.NeutronException):
    message = _("The error occurred when posting to zenic: %(reason)s")

    def __init__(self, **kwargs):
        self.msg = self.message % kwargs


class ZenicQoSPlugin(QoSPlugin):
    def __init__(self):
        conf_proxy = get_rest_proxy_conf()
        if conf_proxy.RESTPROXY.servers is not '':
            self.servers = ZenicRestServerPool(
                conf_proxy.RESTPROXY.servers,
                conf_proxy.RESTPROXY.server_auth,
                conf_proxy.RESTPROXY.server_ssl,
                conf_proxy.RESTPROXY.no_ssl_validation,
                conf_proxy.RESTPROXY.ssl_sticky,
                conf_proxy.RESTPROXY.ssl_cert_directory,
                conf_proxy.RESTPROXY.consistency_interval,
                conf_proxy.RESTPROXY.server_timeout,
                conf_proxy.RESTPROXY.cache_connections,
                conf_proxy.RESTPROXY.base_url)
        super(ZenicQoSPlugin, self).__init__()

    def create_policy(self, context, policy):
        LOG.info(_("create_qos_policy1 policy= %s"), policy)
        policy = super(ZenicQoSPlugin, self).create_policy(context, policy)
        LOG.info(_("create_qos_policy2 policy= %s"), policy)
        try:
            self.servers.create_policy(context, policy)
        except BaseException:
            super(ZenicQoSPlugin, self).delete_policy(context, policy['id'])
            raise PostZenicRestError(reason="create qos policy!")
        return policy

    def update_policy(self, context, policy_id, policy):
        LOG.info(_("update_qos_policy policy=%s"), policy)
        try:
            self.servers.update_policy(context, policy_id, policy)
        except BaseException:
            raise PostZenicRestError(reason="update qos policy!")
        policy = super(ZenicQoSPlugin, self).update_policy(context, policy_id,
                                                           policy)
        return policy

    def delete_policy(self, context, policy_id):
        LOG.info(_("delete_qos_policy id = %s"), policy_id)
        try:
            self.servers.delete_policy(context, policy_id)
        except BaseException:
            raise PostZenicRestError(reason="delete qos policy!")
        super(ZenicQoSPlugin, self).delete_policy(context, policy_id)

    def create_policy_bandwidth_limit_rule(self, context, policy_id,
                                           bandwidth_limit_rule):
        LOG.info(_("create_policy_bandwidth_limit_rule1 rule= %s"),
                 bandwidth_limit_rule)
        rule_cls = rule_object.QosBandwidthLimitRule
        rule = super(ZenicQoSPlugin, self).create_policy_rule(
            context, rule_cls, policy_id, bandwidth_limit_rule)
        LOG.info(_("create_policy_bandwidth_limit_rule2 rule= %s"), rule)
        try:
            self.servers.create_policy_bandwidth_limit_rule(context,
                                                            policy_id, rule)
        except BaseException:
            super(ZenicQoSPlugin, self).delete_policy_rule(
                context, rule_cls, rule['id'], policy_id)
            raise PostZenicRestError(
                reason="create bandwidth limit rule of qos policy!")
        return rule

    def update_policy_bandwidth_limit_rule(self, context, rule_id, policy_id,
                                           bandwidth_limit_rule):
        LOG.info(_("update_policy_bandwidth_limit_rule rule=%s"),
                 bandwidth_limit_rule)
        try:
            self.servers.update_policy_bandwidth_limit_rule(
                context, rule_id, policy_id, bandwidth_limit_rule)
        except BaseException:
            raise PostZenicRestError(
                reason="update bandwidth limit rule of qos policy!")
        rule_cls = rule_object.QosBandwidthLimitRule
        rule = super(ZenicQoSPlugin, self).update_policy_rule(
            context, rule_cls, rule_id, policy_id, bandwidth_limit_rule)
        return rule

    def delete_policy_bandwidth_limit_rule(self, context, rule_id, policy_id):
        LOG.info(_("delete_policy_bandwidth_limit_rule rule_id = %(rule_id)s "
                 "policy_id = %(policy_id)s"), {'rule_id': rule_id,
                                                'policy_id': policy_id})
        try:
            self.servers.delete_policy_bandwidth_limit_rule(context, rule_id,
                                                            policy_id)
        except BaseException:
            raise PostZenicRestError(
                reason="delete bandwidth limit rule of qos policy!")
        rule_cls = rule_object.QosBandwidthLimitRule
        super(ZenicQoSPlugin, self).delete_policy_rule(
            context, rule_cls, rule_id, policy_id)

    def create_policy_dscp_marking_rule(self, context, policy_id,
                                        dscp_marking_rule):
        LOG.info(_("create_policy_dscp_marking_rule1 rule= %s"),
                dscp_marking_rule)
        rule_cls = rule_object.QosDscpMarkingRule
        rule = super(ZenicQoSPlugin, self).create_policy_rule(
            context, rule_cls, policy_id, dscp_marking_rule)
        LOG.info(_("create_policy_dscp_marking_rule2 rule= %s"), rule)
        try:
            self.servers.create_policy_dscp_marking_rule(context,
                                                     policy_id, rule)
        except BaseException:
            super(ZenicQoSPlugin, self).delete_policy_rule(
                context, rule_cls, rule['id'], policy_id)
            raise PostZenicRestError(
                reason="create dscp marking rule of qos policy!")
        return rule

    def update_policy_dscp_marking_rule(self, context, rule_id, policy_id,
                                    dscp_marking_rule):
        LOG.info(_("update_policy_dscp_marking_rule rule=%s"),
                dscp_marking_rule)
        try:
            self.servers.update_policy_dscp_marking_rule(
                context, rule_id, policy_id, dscp_marking_rule)
        except BaseException:
            raise PostZenicRestError(
                reason="update dscp marking rule of qos policy!")
        rule_cls = rule_object.QosDscpMarkingRule
        rule = super(ZenicQoSPlugin, self).update_policy_rule(
            context, rule_cls, rule_id, policy_id, dscp_marking_rule)
        return rule

    def delete_policy_dscp_marking_rule(self, context, rule_id, policy_id):
        LOG.info(_("delete_policy_dscp_marking_rule rule_id = %(rule_id)s "
                   "policy_id = %(policy_id)s"), {'rule_id': rule_id,
                                                  'policy_id': policy_id})
        try:
            self.servers.delete_policy_dscp_marking_rule(context, rule_id,
                                                        policy_id)
        except BaseException:
            raise PostZenicRestError(
                reason="delete dscp marking rule of qos policy!")
        rule_cls = rule_object.QosDscpMarkingRule
        super(ZenicQoSPlugin, self).delete_policy_rule(
            context, rule_cls, rule_id, policy_id)



# Copyright 2017 ZTE, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
'''
Created on 2017年9月21日

@author: 10169294
'''

TAIL_NUM = 5


class CmccUtil(object):

    def filter_gw_port(self, op_version, **kwargs):
        if op_version == "bigcloud_K":
            if kwargs['router']['name'][-TAIL_NUM:] == '_CORE':
                kwargs["gw_port"] = ""
        else:
            return kwargs["gw_port"]

        return kwargs["gw_port"]



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""
This module manages configuration options
"""

from oslo_config import cfg
try:
    from neutron.agent.common import config as agconfig
except Exception:
    from neutron.conf.agent import common as agconfig

try:
    from neutron_lib.utils import net as utils
except Exception:
    from neutron.common import utils
rest_proxy_opts = [
    cfg.ListOpt('servers', default=['localhost:8800'],
                help=_("A comma separated list of Big Switch or Floodlight "
                       "servers and port numbers. The plugin proxies the "
                       "requests to the Big Switch/Floodlight server, "
                       "which performs the networking configuration. Only one"
                       "server is needed per deployment, but you may wish to"
                       "deploy multiple servers to support failover.")),
    cfg.StrOpt('server_auth', default=None, secret=True,
               help=_("The username and password for authenticating against "
                      " the Big Switch or Floodlight controller.")),
    cfg.BoolOpt('server_ssl', default=False,
                help=_("If True, Use SSL when connecting to the Big Switch or "
                       "Floodlight controller.")),
    cfg.BoolOpt('ssl_sticky', default=True,
                help=_("Trust and store the first certificate received for "
                       "each controller address and use it to validate future "
                       "connections to that address.")),
    cfg.BoolOpt('no_ssl_validation', default=False,
                help=_("Disables SSL certificate validation for controllers")),
    cfg.BoolOpt('cache_connections', default=True,
                help=_("Re-use HTTP/HTTPS connections to the controller.")),
    cfg.StrOpt('ssl_cert_directory',
               default='/etc/neutron/plugins/proxyagent/znic/ssl',
               help=_("Directory containing ca_certs and host_certs "
                      "certificate directories.")),
    cfg.BoolOpt('sync_data', default=False,
                help=_("Sync data on connect")),
    cfg.BoolOpt('auto_sync_on_failure', default=True,
                help=_("If neutron fails to create a resource because "
                       "the backend controller doesn't know of a dependency, "
                       "the plugin automatically triggers a full data "
                       "synchronization to the controller.")),
    cfg.IntOpt('consistency_interval', default=60,
               help=_("Time between verifications that the backend controller "
                      "database is consistent with Neutron")),
    cfg.IntOpt('server_timeout', default=10,
               help=_("Maximum number of seconds to wait for proxy request "
                      "to connect and complete.")),
    cfg.IntOpt('thread_pool_size', default=4,
               help=_("Maximum number of threads to spawn to handle large "
                      "volumes of port creations.")),
    cfg.StrOpt('neutron_id', default='neutron-' + utils.get_hostname(),
               deprecated_name='quantum_id',
               help=_("User defined identifier for this Neutron deployment")),
    cfg.BoolOpt('add_meta_server_route', default=True,
                help=_("Flag to decide if a route to the metadata server "
                       "should be injected into the VM")),
    cfg.StrOpt('zenic_version', default="50.1",
               help=_("Version number of the zenic controller corresponding "
                      "to the ml2-plugin")),
    cfg.StrOpt('enable_qos', default="False",
               help=_("Flag to decide if plugin use qos function")),
    cfg.StrOpt('enable_M_qos', default="False",
               help=_("Flag to decide if plugin use M version qos function")),
    cfg.StrOpt('enable_bandwidth', default="False",
               help=_("Flag to decide if plugin use bandwidth function")),
    cfg.StrOpt('enable_pre_commit', default="False",
               help=_("Flag to decide if using pre_commit")),
    cfg.IntOpt('flat_segment_id', default=1000001,
               help=_("Set the only flat network segmentation id for "
                      "zenic controller.")),
    cfg.StrOpt('enable_hierarchical_port', default="False",
               help=_("Flag to decide if using hierarchical port")),
    cfg.StrOpt('vlan_transparent', default="True",
               help=_("Flag to decide if plugin use vlan_transparent "
                      "function.")),
]


def register_config():
    cfg.CONF.register_opts(rest_proxy_opts, "RESTPROXY")
    agconfig.register_root_helper(cfg.CONF)
